@@ -24,7 +24,6 @@
 #include <QJsonDocument>
 #include <QJsonArray>
 #include <kshell.h>
-#include <kconfig.h>
 #include <klocalizedstring.h>
 #include <kaboutdata.h>
 #include <KPluginLoader>
@@ -38,9 +37,11 @@
 #include <QDir>
 #include <QFileInfo>
 #include <QMap>
+#include <QRegularExpression>
 #include <QStandardPaths>
 #include <QStringList>
 #include <QTimer>
+#include <QXmlStreamWriter>
 
 #include <iostream>
 #include <iomanip>
@@ -50,6 +51,48 @@
 #include "../kpackage/private/packagejobthread_p.h"
 
 Q_GLOBAL_STATIC_WITH_ARGS(QTextStream, cout, (stdout))
+Q_GLOBAL_STATIC_WITH_ARGS(QTextStream, cerr, (stderr))
+
+static QVector<KPluginMetaData> listPackageTypes()
+{
+    QStringList libraryPaths;
+
+    const QString subDirectory = QStringLiteral("kpackage/packagestructure");
+
+    Q_FOREACH (const QString &dir, QCoreApplication::libraryPaths()) {
+        QString d = dir + QDir::separator() + subDirectory;
+        if (!d.endsWith(QDir::separator())) {
+            d += QDir::separator();
+        }
+        libraryPaths << d;
+    }
+
+    QVector<KPluginMetaData> offers;
+    Q_FOREACH (const QString &plugindir, libraryPaths) {
+        const QString &_ixfile = plugindir + QStringLiteral("kpluginindex.json");
+        QFile indexFile(_ixfile);
+        if (indexFile.exists()) {
+            indexFile.open(QIODevice::ReadOnly);
+            QJsonDocument jdoc = QJsonDocument::fromBinaryData(indexFile.readAll());
+            indexFile.close();
+
+            QJsonArray plugins = jdoc.array();
+            for (QJsonArray::const_iterator iter = plugins.constBegin(); iter != plugins.constEnd(); ++iter) {
+                const QJsonObject obj = iter->toObject();
+                const QString candidate = obj.value(QStringLiteral("FileName")).toString();
+                offers << KPluginMetaData(obj, candidate);
+            }
+        } else {
+            QVector<KPluginMetaData> plugins = KPluginLoader::findPlugins(plugindir);
+            QVectorIterator<KPluginMetaData> iter(plugins);
+            while (iter.hasNext()) {
+                auto md = iter.next();
+                offers << md;
+            }
+        }
+    }
+    return offers;
+}
 
 namespace KPackage
 {
@@ -91,7 +134,8 @@ void PackageTool::runMain()
         const QString path = d->parser->value(QStringLiteral("hash"));
         KPackage::Package package(structure);
         package.setPath(path);
-        const QString hash = package.contentsHash();
+        const QString hash =
+            QString::fromLocal8Bit(package.cryptographicHash(QCryptographicHash::Sha1));
         if (hash.isEmpty()) {
             d->coutput(i18n("Failed to generate a Package hash for %1", path));
             exit(9);
@@ -109,7 +153,6 @@ void PackageTool::runMain()
     }
 
     QString type = d->parser->value(QStringLiteral("type"));
-    QString packageRoot = type;
     d->pluginTypes.clear();
     d->installer = Package();
 
@@ -121,6 +164,8 @@ void PackageTool::runMain()
         d->package = d->parser->value(QStringLiteral("install"));
     } else if (d->parser->isSet(QStringLiteral("show"))) {
         d->package = d->parser->value(QStringLiteral("show"));
+    } else if (d->parser->isSet(QStringLiteral("appstream-metainfo"))) {
+        d->package = d->parser->value(QStringLiteral("appstream-metainfo"));
     }
 
     if (!QDir::isAbsolutePath(d->package)) {
@@ -143,7 +188,7 @@ void PackageTool::runMain()
 
         if (package.isValid() && package.metadata().isValid()) {
             serviceType = package.metadata().value(QStringLiteral("X-Plasma-ServiceType"));
-            if (serviceType.isEmpty()) {
+            if (serviceType.isEmpty() && !package.metadata().serviceTypes().isEmpty()) {
                 serviceType = package.metadata().serviceTypes().first();
             }
         }
@@ -175,7 +220,10 @@ void PackageTool::runMain()
     if (d->parser->isSet(QStringLiteral("show"))) {
         const QString pluginName = d->package;
         showPackageInfo(pluginName);
-        exit(0);
+        return;
+    } else if (d->parser->isSet(QStringLiteral("appstream-metainfo"))) {
+        const QString pluginName = d->package;
+        showAppstreamInfo(pluginName);
         return;
     }
 
@@ -249,10 +297,7 @@ void PackageTool::runMain()
             }
             QStringList installed = d->packages(d->pluginTypes);
 
-            const QString file = QDir::currentPath() + '/' + pluginName;
-
             if (QFile::exists(d->packageFile)) {
-                const QString file = QDir::currentPath() + '/' + d->package;
                 d->installer.setPath(d->packageFile);
                 if (d->installer.isValid()) {
                     if (d->installer.metadata().isValid()) {
@@ -323,7 +368,7 @@ void PackageTool::showPackageInfo(const QString &pluginName)
 
     KPluginMetaData i = pkg.metadata();
     if (!i.isValid()) {
-        d->coutput(i18n("Error: Can't find plugin metadata: %1", pluginName));
+        *cerr << i18n("Error: Can't find plugin metadata: %1\n", pluginName);
         exit(3);
         return;
     }
@@ -338,8 +383,141 @@ void PackageTool::showPackageInfo(const QString &pluginName)
     exit(0);
 }
 
+bool translateKPluginToAppstream(const QString &tagName, const QString &configField, const QJsonObject &configObject, QXmlStreamWriter& writer, bool canEndWithDot)
+{
+    const QRegularExpression rx(QStringLiteral("%1\\[(.*)\\]").arg(configField));
+    const QJsonValue native = configObject.value(configField);
+    if (native.isUndefined()) {
+        return false;
+    }
+
+    QString content = native.toString();
+    if (!canEndWithDot && content.endsWith(QLatin1Char('.'))) {
+        content.chop(1);
+    }
+    writer.writeTextElement(tagName, content);
+    for (auto it = configObject.begin(), itEnd = configObject.end(); it != itEnd; ++it) {
+        const auto match = rx.match(it.key());
+
+        if (match.hasMatch()) {
+            QString content = it->toString();
+            if (!canEndWithDot && content.endsWith(QLatin1Char('.'))) {
+                content.chop(1);
+            }
+
+            writer.writeStartElement(tagName);
+            writer.writeAttribute("xml:lang", match.captured(1));
+            writer.writeCharacters(content);
+            writer.writeEndElement();
+        }
+    }
+    return true;
+}
+
+void PackageTool::showAppstreamInfo(const QString &pluginName)
+{
+    KPluginMetaData i;
+    KPluginMetaData packageStructureMetaData;
+    //if the path passed is an absolute path, and a metadata file is found under it, use that metadata file to generate the appstream info.
+    // This can happen in the case an application wanting to support kpackage based extensions includes in the same project both the packagestructure plugin and the packages themselves. In that case at build time the packagestructure plugin wouldn't be installed yet
+    if (QFile::exists(pluginName + QStringLiteral("/metadata.desktop"))) {
+        i = KPluginMetaData(pluginName + QStringLiteral("/metadata.desktop"));
+    } else if (QFile::exists(pluginName + QStringLiteral("/metadata.json"))) {
+        i = KPluginMetaData(pluginName + QStringLiteral("/metadata.json"));
+    } else {
+        QString type = QStringLiteral("KPackage/Generic");
+        if (!d->pluginTypes.contains(type) && d->pluginTypes.count() > 0) {
+            type = d->pluginTypes.at(0);
+        }
+        KPackage::Package pkg = KPackage::PackageLoader::self()->loadPackage(type);
+
+        pkg.setDefaultPackageRoot(d->packageRoot);
+
+        if (QFile::exists(d->packageFile)) {
+            pkg.setPath(d->packageFile);
+        } else {
+            pkg.setPath(pluginName);
+        }
+
+        {
+            foreach(const KPluginMetaData &md, listPackageTypes()) {
+                if (md.pluginId() == type) {
+                    packageStructureMetaData = md;
+                    break;
+                }
+            }
+        }
+
+        i = pkg.metadata();
+    }
+
+    if (!i.isValid()) {
+        *cerr << i18n("Error: Can't find plugin metadata: %1\n", pluginName);
+        std::exit(3);
+        return;
+    }
+
+    if (i.rawData().value("NoDisplay").toBool(false)) {
+        *cerr << i18n("%1 has NoDisplay enabled\n", pluginName);
+        std::exit(0);
+    }
+
+    QString parentApp = i.rawData().value("X-KDE-ParentApp").toString();
+    if (parentApp.isEmpty()) {
+        parentApp = packageStructureMetaData.rawData().value("X-KDE-ParentApp").toString();
+    }
+    const QJsonObject rootObject = i.rawData()[QStringLiteral("KPlugin")].toObject();
+
+    QXmlStreamAttributes componentAttributes;
+    if (!parentApp.isEmpty()) {
+        componentAttributes << QXmlStreamAttribute("type", "addon");
+    }
+
+    QXmlStreamWriter writer(cout->device());
+    writer.setAutoFormatting(true);
+    writer.writeStartDocument();
+    writer.writeStartElement("component");
+    writer.writeAttributes(componentAttributes);
+
+    writer.writeTextElement("id", i.pluginId());
+    if (!parentApp.isEmpty()) {
+        writer.writeTextElement("extends", parentApp);
+    }
+    translateKPluginToAppstream("name", "Name", rootObject, writer, false);
+    translateKPluginToAppstream("summary", "Description", rootObject, writer, false);
+    if (!i.website().isEmpty()) {
+        writer.writeStartElement("url");
+        writer.writeAttribute("type", "homepage");
+        writer.writeCharacters(i.website());
+        writer.writeEndElement();
+    }
+    const auto authors = i.authors();
+    if (!authors.isEmpty()) {
+        QStringList authorsText;
+        for (const auto &author: authors) {
+            authorsText += QStringLiteral("%1 <%2>").arg(author.name(), author.emailAddress());
+        }
+        writer.writeTextElement("developer_name", authorsText.join(", "));
+    }
+
+    if (!i.iconName().isEmpty()) {
+        writer.writeStartElement(QStringLiteral("icon"));
+        writer.writeAttribute("type", "stock");
+        writer.writeCharacters(i.iconName());
+        writer.writeEndElement();
+    }
+    writer.writeTextElement("project_license", i.license());
+    writer.writeTextElement("metadata_license", "CC0-1.0");
+    writer.writeEndElement();
+    writer.writeEndDocument();
+
+    exit(0);
+}
+
 QString PackageTool::findPackageRoot(const QString &pluginName, const QString &prefix)
 {
+    Q_UNUSED(pluginName);
+    Q_UNUSED(prefix);
     QString packageRoot;
     if (d->parser->isSet(QStringLiteral("packageroot")) && d->parser->isSet(QStringLiteral("global")) && !d->parser->isSet(QStringLiteral("generate-index"))) {
         qWarning() << i18nc("The user entered conflicting options packageroot and global, this is the error message telling the user he can use only one", "The packageroot and global options conflict with each other, please select only one.");
@@ -349,7 +527,9 @@ QString PackageTool::findPackageRoot(const QString &pluginName, const QString &p
         //qDebug() << "(set via arg) d->packageRoot is: " << d->packageRoot;
     } else if (d->parser->isSet(QStringLiteral("global"))) {
         auto const paths = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, d->packageRoot, QStandardPaths::LocateDirectory);
-        packageRoot = paths.last();
+        if (!paths.isEmpty()) {
+            packageRoot = paths.last();
+        }
     } else {
         packageRoot = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1Char('/') + d->packageRoot;
     }
@@ -422,51 +602,7 @@ void PackageToolPrivate::listTypes()
 
     renderTypeTable(builtIns);
 
-    QStringList libraryPaths;
-
-    const QString subDirectory = QStringLiteral("kpackage/packagestructure");
-
-    Q_FOREACH (const QString &dir, QCoreApplication::libraryPaths()) {
-        QString d = dir + QDir::separator() + subDirectory;
-        if (!d.endsWith(QDir::separator())) {
-            d += QDir::separator();
-        }
-        libraryPaths << d;
-    }
-
-
-    QString pluginFileName;
-
-    QList<KPluginMetaData> offers;
-
-    Q_FOREACH (const QString &plugindir, libraryPaths) {
-        const QString &_ixfile = plugindir + QStringLiteral("kpluginindex.json");
-        QFile indexFile(_ixfile);
-        if (indexFile.exists()) {
-            indexFile.open(QIODevice::ReadOnly);
-            QJsonDocument jdoc = QJsonDocument::fromBinaryData(indexFile.readAll());
-            indexFile.close();
-
-
-            QJsonArray plugins = jdoc.array();
-
-            for (QJsonArray::const_iterator iter = plugins.constBegin(); iter != plugins.constEnd(); ++iter) {
-                const QJsonObject &obj = QJsonValue(*iter).toObject();
-                const QString &candidate = obj.value(QStringLiteral("FileName")).toString();
-                const KPluginMetaData m(obj, candidate);
-                offers << m;
-            }
-        } else {
-            QVector<KPluginMetaData> plugins = KPluginLoader::findPlugins(plugindir);
-            QVectorIterator<KPluginMetaData> iter(plugins);
-            while (iter.hasNext()) {
-                auto md = iter.next();
-                offers << md;
-            }
-        }
-    }
-
-
+    const QVector<KPluginMetaData> offers = listPackageTypes();
 
     if (!offers.isEmpty()) {
         std::cout << std::endl;
@@ -477,10 +613,8 @@ void PackageToolPrivate::listTypes()
             //const QString proot = "";
             KPackage::Package pkg = KPackage::PackageLoader::self()->loadPackage(info.pluginId());
             QString name = info.name();
-            QString comment = info.value(QStringLiteral("Comment"));
             QString plugin = info.pluginId();
             QString path = pkg.defaultPackageRoot();
-            QString types = info.serviceTypes().join(QStringLiteral(","));
             //QString path = defaultPackageRoot;
             plugins.insert(name, QStringList() << plugin << path);
             //qDebug() << "KService stuff:" << name << plugin << comment << path;
@@ -488,23 +622,6 @@ void PackageToolPrivate::listTypes()
 
         renderTypeTable(plugins);
     }
-
-    QStringList desktopFiles = QStandardPaths::locateAll(QStandardPaths::DataLocation, KPACKAGE_RELATIVE_DATA_INSTALL_DIR + QStringLiteral("/packageformats/*rc"), QStandardPaths::LocateFile);
-
-    if (!desktopFiles.isEmpty()) {
-        coutput(i18n("Provided by .desktop files:"));
-        KPackage::PackageStructure structure;
-        QMap<QString, QStringList> plugins;
-        foreach (const QString &file, desktopFiles) {
-            // extract the type
-            KConfig config(file, KConfig::SimpleConfig);
-#pragma message("read config here")
-            // structure.read(&config);
-            // get the name based on the rc file name, just as KPackage::PackageStructure does
-            const QString name = file.left(file.length() - 2);
-            //plugins.insert(name, QStringList() << structure.type() << structure.defaultPackageRoot());
-        }
-    }
 }
 
 void PackageTool::recreateIndex()
@@ -566,4 +683,3 @@ void PackageTool::packageUninstalled(KJob *job)
 } // namespace KPackage
 
 #include "moc_kpackagetool.cpp"
-
